"""Experimental API for the CELLxGENE Discover Census."""

from ._embedding import (
    get_all_available_embeddings,
    get_all_census_versions_with_embedding,
    get_embedding,
    get_embedding_metadata,
    get_embedding_metadata_by_name,
)

__all__ = [
    "get_embedding",
    "get_embedding_metadata",
    "get_embedding_metadata_by_name",
    "get_all_available_embeddings",
    "get_all_census_versions_with_embedding",
]



from typing import Any, Dict, List, Literal

import numpy as np
import pandas as pd
import pytest
import tiledbsoma as soma

import cellxgene_census


@pytest.fixture(scope="session")
def census() -> soma.Collection:
    return cellxgene_census.open_soma(census_version="latest")


@pytest.fixture(scope="session")
def lts_census() -> soma.Collection:
    return cellxgene_census.open_soma(census_version="stable")


@pytest.mark.live_corpus
def test_get_anndata_value_filter(census: soma.Collection) -> None:
    ad = cellxgene_census.get_anndata(
        census,
        organism="Mus musculus",
        obs_value_filter="tissue_general == 'vasculature'",
        var_value_filter="feature_name in ['Gm53058', '0610010K14Rik']",
        column_names={
            "obs": [
                "soma_joinid",
                "cell_type",
                "tissue",
                "tissue_general",
                "assay",
            ],
            "var": ["soma_joinid", "feature_id", "feature_name", "feature_length"],
        },
    )

    assert ad is not None
    assert ad.n_vars == 2
    assert ad.n_obs > 0
    assert (ad.obs.tissue_general == "vasculature").all()
    assert set(ad.var.feature_name) == {"Gm53058", "0610010K14Rik"}


@pytest.mark.live_corpus
def test_get_anndata_coords(census: soma.Collection) -> None:
    ad = cellxgene_census.get_anndata(
        census,
        organism="Mus musculus",
        obs_coords=slice(1000),
        var_coords=slice(2000),
    )

    assert ad is not None
    assert ad.n_vars == 2001
    assert ad.n_obs == 1001


@pytest.mark.live_corpus
def test_get_anndata_allows_missing_obs_or_var_filter(census: soma.Collection) -> None:
    # This test is slightly sensitive to the live data, in that it assumes the
    # existance of certain cell tissue labels and gene feature ids.
    mouse = census["census_data"]["mus_musculus"]

    adata = cellxgene_census.get_anndata(census, organism="Mus musculus", obs_value_filter="tissue == 'aorta'")
    assert adata.n_obs == len(mouse.obs.read(value_filter="tissue == 'aorta'", column_names=["soma_joinid"]).concat())
    assert adata.n_vars == len(mouse.ms["RNA"].var.read(column_names=["soma_joinid"]).concat())

    adata = cellxgene_census.get_anndata(
        census,
        organism="Mus musculus",
        obs_coords=slice(10000),
        var_value_filter="feature_id == 'ENSMUSG00000069581'",
    )
    assert adata.n_obs == 10001
    assert adata.n_vars == 1


@pytest.mark.live_corpus
@pytest.mark.parametrize("layer", ["raw", "normalized"])
def test_get_anndata_x_layer(census: soma.Collection, layer: str) -> None:
    ad = cellxgene_census.get_anndata(
        census,
        organism="Homo sapiens",
        X_name=layer,
        obs_coords=slice(100),
        var_coords=slice(200),
    )

    assert ad.X.shape == (101, 201)
    assert len(ad.layers) == 0


@pytest.mark.live_corpus
@pytest.mark.parametrize("layers", [["raw", "normalized"], ["normalized", "raw"]])
def test_get_anndata_two_layers(census: soma.Collection, layers: List[str]) -> None:
    ad_primary_layer_in_X = cellxgene_census.get_anndata(
        census,
        organism="Homo sapiens",
        X_name=layers[0],
        obs_coords=slice(100),
        var_coords=slice(200),
    )

    ad_secondary_layer_in_X = cellxgene_census.get_anndata(
        census,
        organism="Homo sapiens",
        X_name=layers[1],
        obs_coords=slice(100),
        var_coords=slice(200),
    )

    ad_multiple_layers = cellxgene_census.get_anndata(
        census,
        organism="Homo sapiens",
        X_name=layers[0],
        X_layers=[layers[1]],
        obs_coords=slice(100),
        var_coords=slice(200),
    )

    assert layers[1] in ad_multiple_layers.layers
    assert ad_multiple_layers.X.shape == (101, 201)
    assert ad_multiple_layers.layers[layers[1]].shape == (101, 201)

    # Assert that matrices of multilayer anndata are equal to one-layer-at-time anndatas
    assert np.array_equal(ad_multiple_layers.X.data, ad_primary_layer_in_X.X.data)
    assert np.array_equal(ad_multiple_layers.layers[layers[1]].data, ad_secondary_layer_in_X.X.data)


@pytest.mark.live_corpus
def test_get_anndata_wrong_layer_names(census: soma.Collection) -> None:
    with pytest.raises(ValueError) as raise_info:
        cellxgene_census.get_anndata(
            census,
            organism="Homo sapiens",
            X_name="this_layer_name_is_bad",
            obs_coords=slice(100),
            var_coords=slice(200),
        )

        assert raise_info.value.args[0] == "Unknown X layer name"

    with pytest.raises(ValueError) as raise_info:
        cellxgene_census.get_anndata(
            census,
            organism="Homo sapiens",
            X_name="raw",
            X_layers=["this_layer_name_is_bad"],
            obs_coords=slice(100),
            var_coords=slice(200),
        )

        assert raise_info.value.args[0] == "Unknown X layer name"


@pytest.mark.live_corpus
@pytest.mark.parametrize("obsm_layer", ["scvi", "geneformer"])
def test_get_anndata_obsm_one_layer(lts_census: soma.Collection, obsm_layer: str) -> None:
    # NOTE: this test will break after next LTS release (>2023-12-15), since scvi and geneformer
    # won't be distributed as part of `obsm_layers` anymore. Delete this test when it happens.
    ad = cellxgene_census.get_anndata(
        lts_census,
        organism="Homo sapiens",
        X_name="raw",
        obs_coords=slice(100),
        var_coords=slice(200),
        obsm_layers=[obsm_layer],
    )

    assert len(ad.obsm.keys()) == 1
    assert obsm_layer in ad.obsm.keys()
    assert ad.obsm[obsm_layer].shape[0] == 101


@pytest.mark.live_corpus
@pytest.mark.parametrize("obsm_layers", [["scvi", "geneformer"]])
def test_get_anndata_obsm_two_layers(lts_census: soma.Collection, obsm_layers: List[str]) -> None:
    # NOTE: this test will break after next LTS release (>2023-12-15), since scvi and geneformer
    # won't be distributed as part of `obsm_layers` anymore. Delete this test when it happens.
    ad = cellxgene_census.get_anndata(
        lts_census,
        organism="Homo sapiens",
        X_name="raw",
        obs_coords=slice(100),
        var_coords=slice(200),
        obsm_layers=obsm_layers,
    )

    assert len(ad.obsm.keys()) == 2
    for obsm_layer in obsm_layers:
        assert obsm_layer in ad.obsm.keys()
        assert ad.obsm[obsm_layer].shape[0] == 101


@pytest.mark.live_corpus
@pytest.mark.parametrize("obs_embeddings", [["scvi", "geneformer", "uce"]])
def test_get_anndata_obs_embeddings(lts_census: soma.Collection, obs_embeddings: List[str]) -> None:
    # NOTE: when the next LTS gets released (>2023-12-15), embeddings may or may not be available,
    # so this test could require adjustments.
    ad = cellxgene_census.get_anndata(
        lts_census,
        organism="Homo sapiens",
        X_name="raw",
        obs_coords=slice(100),
        var_coords=slice(200),
        obs_embeddings=obs_embeddings,
    )

    assert len(ad.obsm.keys()) == 3
    assert len(ad.varm.keys()) == 0
    for obsm_layer in obs_embeddings:
        assert obsm_layer in ad.obsm.keys()
        assert ad.obsm[obsm_layer].shape[0] == 101


@pytest.mark.live_corpus
@pytest.mark.parametrize("var_embeddings", [["nmf"]])
def test_get_anndata_var_embeddings(lts_census: soma.Collection, var_embeddings: List[str]) -> None:
    # NOTE: when the next LTS gets released (>2023-12-15), embeddings may or may not be available,
    # so this test could require adjustments.

    ad = cellxgene_census.get_anndata(
        lts_census,
        organism="Homo sapiens",
        X_name="raw",
        obs_coords=slice(100),
        var_coords=slice(200),
        var_embeddings=var_embeddings,
    )

    assert len(ad.obsm.keys()) == 0
    assert len(ad.varm.keys()) == 1
    for varm_layers in var_embeddings:
        assert varm_layers in ad.varm.keys()
        assert ad.varm[varm_layers].shape[0] == 201


@pytest.mark.live_corpus
def test_get_anndata_obsm_layers_and_add_obs_embedding_fails(lts_census: soma.Collection) -> None:
    """Fails if both `obsm_layers` and `obs_embeddings` are specified."""
    with pytest.raises(ValueError):
        cellxgene_census.get_anndata(
            lts_census,
            organism="Homo sapiens",
            X_name="raw",
            obs_coords=slice(100),
            var_coords=slice(200),
            obsm_layers=["scvi"],
            obs_embeddings=["scvi"],
        )


def _map_to_get_anndata_args(query: Dict[str, Any], axis: Literal["obs", "var"]) -> Dict[str, Any]:
    """Helper to map arguments of get_obs/ get_var to get_anndata."""
    result = {}
    if "coords" in query:
        result[f"{axis}_coords"] = query["coords"]
    if "value_filter" in query:
        result[f"{axis}_value_filter"] = query["value_filter"]
    if "column_names" in query:
        result["column_names"] = {axis: query["column_names"]}
    return result


@pytest.mark.live_corpus
@pytest.mark.parametrize(
    "query",
    [
        pytest.param(
            {
                "coords": slice(100),
                "column_names": [
                    "soma_joinid",
                    "cell_type",
                    "tissue",
                    "tissue_general",
                    "assay",
                ],
            },
            id="coords+column-names",
        ),
        pytest.param({"coords": slice(100, 300)}, id="coords"),
        pytest.param({"value_filter": "tissue_general == 'vasculature'"}, id="value_filter"),
    ],
)
def test_get_obs(lts_census: soma.Collection, query: Dict[str, Any]) -> None:
    adata_obs = cellxgene_census.get_anndata(
        lts_census, organism="Mus musculus", **_map_to_get_anndata_args(query, "obs")
    ).obs
    only_obs = cellxgene_census.get_obs(lts_census, "Mus musculus", **query)
    # account for a difference:
    only_obs.index = only_obs.index.astype(str)

    pd.testing.assert_frame_equal(adata_obs, only_obs)


@pytest.mark.live_corpus
@pytest.mark.parametrize(
    "query",
    [
        pytest.param(
            {
                "coords": slice(100),
                "column_names": ["soma_joinid", "feature_id", "feature_name", "feature_length"],
            },
            id="coords+column-names",
        ),
        pytest.param({"coords": slice(100, 300)}, id="coords"),
        pytest.param({"value_filter": "feature_name in ['Gm53058', '0610010K14Rik']"}, id="value_filter"),
    ],
)
def test_get_var(lts_census: soma.Collection, query: Dict[str, Any]) -> None:
    adata_var = cellxgene_census.get_anndata(
        lts_census, organism="Mus musculus", obs_coords=slice(0), **_map_to_get_anndata_args(query, "var")
    ).var
    only_var = cellxgene_census.get_var(lts_census, "Mus musculus", **query)
    # AnnData instantiation converts the index to string, so we match that behaviour for comparisons sake
    only_var.index = only_var.index.astype(str)

    pd.testing.assert_frame_equal(adata_var, only_var)



import numpy as np
import numpy.ma as ma
import numpy.typing as npt
import pytest
from scipy import sparse

from cellxgene_census.experimental.pp._online import (
    CountsAccumulator,
    MeanAccumulator,
    MeanVarianceAccumulator,
)


def allclose(a: npt.NDArray[np.float64], b: npt.NDArray[np.float64]) -> bool:
    return np.allclose(a, b, atol=1e-5, rtol=1e-2, equal_nan=True)


@pytest.fixture
def matrix(m: int, n: int) -> sparse.coo_matrix:
    m = 100 * sparse.random(
        m,
        n,
        density=0.1,
        format="coo",
        dtype=np.float32,
        random_state=np.random.default_rng(),
    )
    m.row.flags.writeable = False  # type: ignore[attr-defined]
    m.col.flags.writeable = False  # type: ignore[attr-defined]
    m.data.flags.writeable = False  # type: ignore[attr-defined]
    return m


@pytest.mark.experimental
@pytest.mark.parametrize("stride", [101, 53])
@pytest.mark.parametrize("n_batches", [1, 3, 11, 101])
@pytest.mark.parametrize("m,n", [(1200, 511), (100001, 57)])
@pytest.mark.parametrize("ddof", [0, 1, 100])
@pytest.mark.parametrize("nnz_only", [True, False])
def test_meanvar(matrix: sparse.coo_matrix, n_batches: int, stride: int, ddof: int, nnz_only: bool) -> None:
    rng = np.random.default_rng()
    batches_prob = rng.random(n_batches)
    batches_prob /= batches_prob.sum()
    batches = rng.choice(n_batches, matrix.shape[0], p=batches_prob)
    batches.flags.writeable = False

    batch_id, batch_count = np.unique(batches, return_counts=True)
    n_samples = np.zeros((n_batches,), dtype=np.int64)
    n_samples[batch_id] = batch_count
    assert n_samples.sum() == matrix.shape[0]
    assert len(n_samples) == n_batches

    # nnz_only only if there is a single batch
    should_nnz_only = nnz_only and n_batches == 1

    olmv = MeanVarianceAccumulator(n_batches, n_samples, matrix.shape[1], ddof, nnz_only=should_nnz_only)
    for i in range(0, matrix.nnz, stride):
        batch_vec = batches[matrix.row[i : i + stride]] if n_batches > 1 else None
        olmv.update(matrix.col[i : i + stride], matrix.data[i : i + stride], batch_vec)
    batches_u, batches_var, all_u, all_var = olmv.finalize()

    assert isinstance(all_u, np.ndarray)
    assert isinstance(all_var, np.ndarray)
    assert isinstance(batches_u, np.ndarray)
    assert isinstance(batches_var, np.ndarray)

    assert all_u.shape == (matrix.shape[1],)
    assert all_var.shape == (matrix.shape[1],)
    assert batches_u.shape == (n_batches, matrix.shape[1])
    assert batches_var.shape == (n_batches, matrix.shape[1])

    dense = matrix.toarray()

    if should_nnz_only:
        mask = np.ones(matrix.shape)
        r, c = matrix.tolil().nonzero()
        for x, y in zip(r, c):
            mask[x, y] = 0
        masked: np.ma.MaskedArray[np.bool_, np.dtype[np.float64]] = ma.masked_array(dense, mask=mask)  # type: ignore[no-untyped-call]
        mean = masked.mean(axis=0)  # type: ignore[no-untyped-call]
        assert allclose(all_u, mean)

        nv = masked.var(axis=0, ddof=ddof)  # type: ignore[no-untyped-call]
        assert allclose(all_var, nv)

    else:
        assert allclose(all_u, dense.mean(axis=0))
        assert allclose(all_var, dense.var(axis=0, ddof=ddof, dtype=np.float64))

        matrix = matrix.tocsr()  # COO not conveniently indexable
        for batch in range(n_batches):
            dense = matrix[batches == batch, :].toarray()
            assert allclose(batches_u[batch], dense.mean(axis=0))
            assert allclose(batches_var[batch], dense.var(axis=0, ddof=ddof, dtype=np.float64))


@pytest.mark.experimental
@pytest.mark.parametrize("m,n", [(1200, 511), (100001, 57)])
def test_meanvar_nnz_only_batches_fails(matrix: sparse.coo_matrix) -> None:
    n_batches = 10
    nnz_only = True
    n_samples = np.zeros((n_batches,), dtype=np.int64)
    ddof = 1
    with pytest.raises(ValueError):
        MeanVarianceAccumulator(n_batches, n_samples, matrix.shape[1], ddof, nnz_only=nnz_only)


@pytest.mark.experimental
@pytest.mark.parametrize("stride", [101, 53])
@pytest.mark.parametrize("m,n", [(1200, 511), (100001, 57)])
def test_mean(matrix: sparse.coo_matrix, stride: int) -> None:
    m_acc = MeanAccumulator(matrix.shape[0], matrix.shape[1])
    for i in range(0, matrix.nnz, stride):
        m_acc.update(matrix.col[i : i + stride], matrix.data[i : i + stride])
    u = m_acc.finalize()

    assert isinstance(u, np.ndarray)
    assert u.shape == (matrix.shape[1],)

    dense = matrix.toarray()
    assert allclose(u, dense.mean(axis=0))


@pytest.mark.experimental
@pytest.mark.parametrize("stride", [101, 53])
@pytest.mark.parametrize("n_batches", [1, 3, 11, 101])
@pytest.mark.parametrize("m,n", [(1200, 511), (100001, 57)])
def test_counts(matrix: sparse.coo_matrix, n_batches: int, stride: int) -> None:
    rng = np.random.default_rng()
    batches_prob = rng.random(n_batches)
    batches_prob /= batches_prob.sum()
    batches = rng.choice(n_batches, matrix.shape[0], p=batches_prob)
    batches.flags.writeable = False

    batch_id, batch_count = np.unique(batches, return_counts=True)
    n_samples = np.zeros((n_batches,), dtype=np.int64)
    n_samples[batch_id] = batch_count
    assert n_samples.sum() == matrix.shape[0]
    assert len(n_samples) == n_batches

    clip_val = 50 * np.random.rand(n_batches, matrix.shape[1])

    ca = CountsAccumulator(n_batches, matrix.shape[1], clip_val)
    for i in range(0, matrix.nnz, stride):
        batch_vec = batches[matrix.row[i : i + stride]] if n_batches > 1 else None
        ca.update(matrix.col[i : i + stride], matrix.data[i : i + stride], batch_vec)
    counts_sum, counts_squared_sum = ca.finalize()

    assert isinstance(counts_sum, np.ndarray)
    assert isinstance(counts_squared_sum, np.ndarray)

    assert counts_sum.shape == (n_batches, matrix.shape[1])
    assert counts_squared_sum.shape == (n_batches, matrix.shape[1])

    matrix = matrix.tocsr()  # COO not conveniently indexable
    for batch in range(n_batches):
        dense = matrix[batches == batch, :].toarray()
        assert allclose(counts_sum[batch], np.minimum(dense, clip_val[batch]).sum(axis=0))
        assert allclose(
            counts_squared_sum[batch],
            (np.minimum(dense, clip_val[batch]) ** 2).sum(axis=0),
        )


@pytest.mark.experimental
def test_mean_fails_no_variables_or_samples() -> None:
    with pytest.raises(ValueError, match=r"No samples provided - can't calculate mean."):
        MeanAccumulator(n_samples=0, n_variables=100)
    with pytest.raises(ValueError, match=r"No variables provided - can't calculate mean."):
        MeanAccumulator(n_samples=1000, n_variables=0)



# Copyright (c) 2022, Chan Zuckerberg Initiative
#
# Licensed under the MIT License.

"""Methods to support simplifed access to community contributed embeddings."""

from __future__ import annotations

import json
import warnings
from typing import Any, Dict, cast

import numpy as np
import numpy.typing as npt
import pandas as pd
import pyarrow as pa
import requests
import tiledbsoma as soma

from .._open import get_default_soma_context, open_soma
from .._release_directory import CensusVersionDescription, CensusVersionName, get_census_version_directory

CELL_CENSUS_EMBEDDINGS_MANIFEST_URL = "https://contrib.cellxgene.cziscience.com/contrib/cell-census/contributions.json"


def get_embedding_metadata(embedding_uri: str, context: soma.options.SOMATileDBContext | None = None) -> dict[str, Any]:
    """Read embedding metadata and return as a Python dict.

    Args:
        embedding_uri:
            The embedding URI.
        context:
            A custom :class:`tiledbsoma.SOMATileDBContext` which will be used to open the SOMA object. Optional,
            defaults to ``None``.

    Returns:
        A Python dictionary containing metadata describing the embedding.

    Examples:
        >>> get_experiment_metadata(uri)

    """
    # Allow the user to override context for exceptional cases (e.g. the aws region)
    context = context or get_default_soma_context()

    with soma.open(embedding_uri, context=context) as E:
        # read embedding metadata and decode the JSON-encoded string
        embedding_metadata = json.loads(E.metadata["CxG_embedding_info"])
        assert isinstance(embedding_metadata, dict)

    return cast(Dict[str, Any], embedding_metadata)


def _get_embedding(
    census: soma.Collection,
    census_directory: dict[CensusVersionName, CensusVersionDescription],
    census_version: str,
    embedding_uri: str,
    obs_soma_joinids: npt.NDArray[np.int64] | pa.Array,
    context: soma.options.SOMATileDBContext | None = None,
) -> npt.NDArray[np.float32]:
    """Private. Like get_embedding, but accepts a Census object and a Census directory."""
    if isinstance(obs_soma_joinids, (pa.Array, pa.ChunkedArray, pd.Series)):
        obs_soma_joinids = obs_soma_joinids.to_numpy()
    assert isinstance(obs_soma_joinids, np.ndarray)
    if obs_soma_joinids.dtype != np.int64:
        raise TypeError("obs_soma_joinids must be array of int64")

    # Allow the user to override context for exceptional cases (e.g. the aws region)
    context = context or get_default_soma_context()

    # Attempt to resolve census version aliases
    resolved_census_version = census_directory.get(census_version, None)

    with soma.open(embedding_uri, context=context) as E:
        embedding_metadata = json.loads(E.metadata["CxG_embedding_info"])

        if resolved_census_version is None:
            warnings.warn(
                "Unable to determine Census version - skipping validation of Census and embedding version.",
                stacklevel=1,
            )
        elif resolved_census_version != census_directory.get(embedding_metadata["census_version"], None):
            raise ValueError("Census and embedding mismatch - census_version not equal")

        with open_soma(census_version=census_version, context=context) as census:
            experiment_name = embedding_metadata["experiment_name"]
            if experiment_name not in census["census_data"]:
                raise ValueError("Census and embedding mismatch - experiment_name does not exist")
            measurement_name = embedding_metadata["measurement_name"]
            if measurement_name not in census["census_data"][experiment_name].ms:
                raise ValueError("Census and embedding mismatch - measurement_name does not exist")

        embedding_shape = (len(obs_soma_joinids), E.shape[1])
        embedding = np.full(embedding_shape, np.NaN, dtype=np.float32, order="C")

        obs_indexer = soma.IntIndexer(obs_soma_joinids, context=E.context)
        for tbl in E.read(coords=(obs_soma_joinids,)).tables():
            obs_idx = obs_indexer.get_indexer(tbl.column("soma_dim_0").to_numpy())
            feat_idx = tbl.column("soma_dim_1").to_numpy()
            emb = tbl.column("soma_data")

            indices = obs_idx * E.shape[1] + feat_idx
            np.put(embedding.reshape(-1), indices, emb)

    return embedding


def get_embedding(
    census_version: str,
    embedding_uri: str,
    obs_soma_joinids: npt.NDArray[np.int64] | pa.Array,
    context: soma.options.SOMATileDBContext | None = None,
) -> npt.NDArray[np.float32]:
    """Read cell (obs) embeddings and return as a dense :class:`numpy.ndarray`. Any cells without
    an embedding will return NaN values.

    Args:
        census_version:
            The Census version tag, e.g., ``"2023-12-15"``. Used to verify that the contents of
            the embedding contain embedded cells from the same Census version.
        embedding_uri:
            The URI containing the embedding data.
        obs_soma_joinids:
            The slice of the embedding to fetch and return.
        context:
            A custom :class:`tiledbsoma.SOMATileDBContext` which will be used to open the SOMA object.
            Optional, defaults to ``None``.

    Returns:
        A :class:`numpy.ndarray` containing the embeddings. Embeddings are positionally
        indexed by the ``obs_soma_joinids``. In other words, the cell identified by
        ``obs_soma_joinids[i]`` corresponds to the ``ith`` position in the returned
        :class:`numpy.ndarray`.

    Raises:
        ValueError: if the Census and embedding are mismatched.

    Lifecycle:
        experimental

    Examples:
        >>> obs_somaids_to_fetch = np.array([10,11], dtype=np.int64)
        >>> emb = cellxgene_census.experimental.get_embedding('2023-12-15', embedding_uri, obs_somaids_to_fetch)
        >>> emb.shape
        (2, 200)
        >>> emb[:, 0:4]
        array([[ 0.02954102,  1.0390625 , -0.14550781, -0.40820312],
            [-0.00224304,  1.265625  ,  0.05883789, -0.7890625 ]],
            dtype=float32)

    """
    census_directory = get_census_version_directory()

    with open_soma(census_version=census_version, context=context) as census:
        return _get_embedding(
            census, census_directory, census_version, embedding_uri, obs_soma_joinids, context=context
        )


def get_embedding_metadata_by_name(
    embedding_name: str, organism: str, census_version: str, embedding_type: str | None = "obs_embedding"
) -> dict[str, Any]:
    """Return metadata for a specific embedding. If more embeddings match the query parameters,
    the most recent one will be returned.

    Args:
        embedding_name:
            The name of the embedding, e.g. "scvi".
        organism:
            The organism for which the embedding is associated.
        census_version:
            The Census version tag, e.g., ``"2023-12-15"``.
        embedding_type:
            Either "obs_embedding" or "var_embedding". Defaults to "obs_embedding".

    Returns:
        A dictionary containing metadata describing the embedding.

    Raises:
        ValueError: if no embeddings are found for the specified query parameters.

    """
    response = requests.get(CELL_CENSUS_EMBEDDINGS_MANIFEST_URL)
    response.raise_for_status()

    manifest = cast(Dict[str, Dict[str, Any]], response.json())
    embeddings = []
    for _, obj in manifest.items():
        if (
            obj["embedding_name"] == embedding_name
            and obj["experiment_name"] == organism
            and obj["data_type"] == embedding_type
            and obj["census_version"] == census_version
        ):
            embeddings.append(obj)

    if len(embeddings) == 0:
        raise ValueError(f"No embeddings found for {embedding_name}, {organism}, {census_version}, {embedding_type}")

    return sorted(embeddings, key=lambda x: x["submission_date"])[-1]


def get_all_available_embeddings(census_version: str) -> list[dict[str, Any]]:
    """Return a dictionary of all available embeddings for a given Census version.

    Args:
        census_version:
            The Census version tag, e.g., ``"2023-12-15"``.

    Returns:
        A list of dictionaries, each containing metadata describing an available embedding.

    Examples:
        >>> get_all_available_embeddings('2023-12-15')
        [{
            'experiment_name': 'experiment_1',
            'measurement_name': 'RNA',
            'organism': "homo_sapiens",
            'census_version': '2023-12-15',
            'n_embeddings': 1000,
            'n_features': 200,
            'uri': 's3://bucket/embedding_1'
        }]

    """
    response = requests.get(CELL_CENSUS_EMBEDDINGS_MANIFEST_URL)
    response.raise_for_status()

    embeddings = []
    manifest = response.json()
    for _, obj in manifest.items():
        if obj["census_version"] == census_version:
            embeddings.append(obj)

    return embeddings


def get_all_census_versions_with_embedding(
    embedding_name: str, organism: str, embedding_type: str | None = "obs_embedding"
) -> list[str]:
    """Get a list of all census versions that contain a specific embedding.

    Args:
        embedding_name:
            The name of the embedding, e.g. "scvi".
        organism:
            The organism for which the embedding is associated.
        embedding_type:
            The type of embedding. Defaults to "obs_embedding".

    Returns:
        A list of census versions that contain the specified embedding.
    """
    response = requests.get(CELL_CENSUS_EMBEDDINGS_MANIFEST_URL)
    response.raise_for_status()

    manifest = response.json()
    return sorted(
        {
            obj["census_version"]
            for obj in manifest.values()
            if obj["embedding_name"] == embedding_name
            and obj["experiment_name"] == organism
            and obj["data_type"] == embedding_type
        }
    )



# Copyright (c) 2022, Chan Zuckerberg Initiative
#
# Licensed under the MIT License.

"""Presence matrix methods.

Methods to retrieve the feature dataset presence matrix.
"""

import tiledbsoma as soma
from scipy import sparse

from ._experiment import _get_experiment


def get_presence_matrix(
    census: soma.Collection,
    organism: str,
    measurement_name: str = "RNA",
) -> sparse.csr_matrix:
    """Read the feature dataset presence matrix and return as a :class:`scipy.sparse.csr_array`. The
    returned sparse matrix is indexed on the first dimension by the dataset ``soma_joinid`` values,
    and on the second dimension by the ``var`` :class:`pandas.DataFrame` ``soma_joinid`` values.

    Args:
        census:
            The census from which to read the presence matrix.
        organism:
            The organism to query, usually one of ``"Homo sapiens"`` or ``"Mus musculus"``.
        measurement_name:
            The measurement object to query. Deafults to ``"RNA"``.

    Returns:
        A :class:`scipy.sparse.csr_array` object containing the presence matrix.

    Raises:
        ValueError: if the organism cannot be found.

    Lifecycle:
        maturing

    Examples:
        >>> get_presence_matrix(census, "Homo sapiens", "RNA")
        <321x60554 sparse array of type '<class 'numpy.uint8'>'
        with 6441269 stored elements in Compressed Sparse Row format>
    """
    exp = _get_experiment(census, organism)
    presence = exp.ms[measurement_name]["feature_dataset_presence_matrix"]
    return presence.read((slice(None),)).coos().concat().to_scipy().tocsr()



import urllib.parse

import tiledbsoma as soma


def _uri_join(base: str, url: str) -> str:
    """Like urllib.parse.urljoin, but doesn't get confused by s3://."""
    p_url = urllib.parse.urlparse(url)
    if p_url.netloc:
        return url

    p_base = urllib.parse.urlparse(base)
    path = urllib.parse.urljoin(p_base.path, p_url.path)
    parts = [
        p_base.scheme,
        p_base.netloc,
        path,
        p_url.params,
        p_url.query,
        p_url.fragment,
    ]
    return urllib.parse.urlunparse(parts)


def _extract_census_version(census: soma.Collection) -> str:
    """Extract the Census version from the given Census object."""
    try:
        version: str = urllib.parse.urlparse(census.uri).path.split("/")[2]
    except (KeyError, IndexError):
        raise ValueError("Unable to extract Census version.") from None

    return version



import pytest
import requests_mock as rm

from cellxgene_census.experimental import (
    get_all_available_embeddings,
    get_all_census_versions_with_embedding,
    get_embedding_metadata_by_name,
)
from cellxgene_census.experimental._embedding import CELL_CENSUS_EMBEDDINGS_MANIFEST_URL


def test_get_embedding_metadata_by_name(requests_mock: rm.Mocker) -> None:
    mock_embeddings = {
        "embedding-id-1": {
            "id": "embedding-id-1",
            "embedding_name": "emb_1",
            "title": "Embedding 1",
            "description": "First embedding",
            "experiment_name": "homo_sapiens",
            "data_type": "obs_embedding",
            "census_version": "2023-12-15",
            "submission_date": "2023-11-15",
        },
        "embedding-id-2": {
            "id": "embedding-id-2",
            "embedding_name": "emb_1",
            "title": "Embedding 2",
            "description": "Second embedding",
            "experiment_name": "homo_sapiens",
            "data_type": "obs_embedding",
            "census_version": "2023-12-15",
            "submission_date": "2023-12-31",
        },
        "embedding-id-3": {
            "id": "embedding-id-3",
            "embedding_name": "emb_3",
            "title": "Embedding 3",
            "description": "Third embedding",
            "experiment_name": "homo_sapiens",
            "data_type": "obs_embedding",
            "census_version": "2023-12-15",
            "submission_date": "2023-11-15",
        },
    }
    requests_mock.real_http = True
    requests_mock.get(CELL_CENSUS_EMBEDDINGS_MANIFEST_URL, json=mock_embeddings)

    embedding = get_embedding_metadata_by_name(
        "emb_1", organism="homo_sapiens", census_version="2023-12-15", embedding_type="obs_embedding"
    )
    assert embedding is not None
    assert embedding["id"] == "embedding-id-2"  # most recent version
    assert embedding == mock_embeddings["embedding-id-2"]

    embedding = get_embedding_metadata_by_name(
        "emb_3", organism="homo_sapiens", census_version="2023-12-15", embedding_type="obs_embedding"
    )
    assert embedding is not None
    assert embedding["id"] == "embedding-id-3"
    assert embedding == mock_embeddings["embedding-id-3"]

    with pytest.raises(ValueError):
        get_embedding_metadata_by_name(
            "emb_2", organism="homo_sapiens", census_version="2023-12-15", embedding_type="obs_embedding"
        )
        get_embedding_metadata_by_name(
            "emb_1", organism="mus_musculus", census_version="2023-12-15", embedding_type="obs_embedding"
        )
        get_embedding_metadata_by_name(
            "emb_1", organism="homo_sapiens", census_version="2023-10-15", embedding_type="obs_embedding"
        )
        get_embedding_metadata_by_name(
            "emb_1", organism="mus_musculus", census_version="2023-12-15", embedding_type="var_embedding"
        )


def test_get_all_available_embeddings(requests_mock: rm.Mocker) -> None:
    mock_embeddings = {
        "embedding-id-1": {
            "id": "embedding-id-1",
            "embedding_name": "emb_1",
            "title": "Embedding 1",
            "description": "First embedding",
            "experiment_name": "homo_sapiens",
            "measurement_name": "RNA",
            "n_embeddings": 1000,
            "n_features": 200,
            "data_type": "obs_embedding",
            "census_version": "2023-12-15",
        },
        "embedding-id-2": {
            "id": "embedding-id-2",
            "embedding_name": "emb_2",
            "title": "Embedding 2",
            "description": "Second embedding",
            "experiment_name": "homo_sapiens",
            "measurement_name": "RNA",
            "n_embeddings": 1000,
            "n_features": 200,
            "data_type": "obs_embedding",
            "census_version": "2023-12-15",
        },
    }
    requests_mock.real_http = True
    requests_mock.get(CELL_CENSUS_EMBEDDINGS_MANIFEST_URL, json=mock_embeddings)

    embeddings = get_all_available_embeddings("2023-12-15")
    assert embeddings is not None
    assert len(embeddings) == 2

    # Query for a non existing version of the Census
    embeddings = get_all_available_embeddings("2024-12-15")
    assert len(embeddings) == 0


def test_get_all_census_versions_with_embedding(requests_mock: rm.Mocker) -> None:
    mock_embeddings = {
        "embedding-id-1": {
            "id": "embedding-id-1",
            "embedding_name": "emb_1",
            "title": "Embedding 1",
            "description": "First embedding",
            "experiment_name": "homo_sapiens",
            "data_type": "obs_embedding",
            "census_version": "2023-12-15",
        },
        "embedding-id-2": {
            "id": "embedding-id-2",
            "embedding_name": "emb_1",
            "title": "Embedding 2",
            "description": "Second embedding",
            "experiment_name": "homo_sapiens",
            "data_type": "obs_embedding",
            "census_version": "2023-12-15",
        },
        "embedding-id-3": {
            "id": "embedding-id-3",
            "embedding_name": "emb_1",
            "title": "Embedding 3",
            "description": "Third embedding",
            "experiment_name": "mus_musculus",
            "data_type": "obs_embedding",
            "census_version": "2023-12-15",
        },
        "embedding-id-4": {
            "id": "embedding-id-4",
            "embedding_name": "emb_1",
            "title": "Embedding 4",
            "description": "Fourth embedding",
            "experiment_name": "mus_musculus",
            "data_type": "obs_embedding",
            "census_version": "2024-01-01",
        },
        "embedding-id-5": {
            "id": "embedding-id-5",
            "embedding_name": "emb_2",
            "title": "Embedding 5",
            "description": "Fifth embedding",
            "experiment_name": "mus_musculus",
            "data_type": "var_embedding",
            "census_version": "2023-12-15",
        },
    }
    requests_mock.real_http = True
    requests_mock.get(CELL_CENSUS_EMBEDDINGS_MANIFEST_URL, json=mock_embeddings)

    versions = get_all_census_versions_with_embedding("emb_1", organism="homo_sapiens", embedding_type="obs_embedding")
    assert versions == ["2023-12-15"]

    versions = get_all_census_versions_with_embedding("emb_1", organism="mus_musculus", embedding_type="obs_embedding")
    assert versions == ["2023-12-15", "2024-01-01"]

    versions = get_all_census_versions_with_embedding("emb_1", organism="mus_musculus", embedding_type="var_embedding")
    assert versions == []

    versions = get_all_census_versions_with_embedding("emb_2", organism="mus_musculus", embedding_type="var_embedding")
    assert versions == ["2023-12-15"]



# Copyright (c) 2022, Chan Zuckerberg Initiative
#
# Licensed under the MIT License.

"""Experiments handler.

Contains methods to retrieve SOMA Experiments.
"""

import re

import tiledbsoma as soma


def _get_experiment_name(organism: str) -> str:
    """Given an organism name, return the experiment name."""
    # lower/snake case the organism name to find the experiment name
    return re.sub(r"[ ]+", "_", organism).lower()


def _get_experiment(census: soma.Collection, organism: str) -> soma.Experiment:
    """Given a census :class:`tiledbsoma.Collection`, return the experiment for the named organism.
    Organism matching is somewhat flexible, attempting to map from human-friendly
    names to the underlying collection element name.

    Args:
        census:
            The census.
        organism:
            The organism name, e.g., ``"Homo sapiens"``.

    Returns:
        An :class:`tiledbsoma.Experiment` object with the requested experiment.

    Raises:
        ValueError: if unable to find the specified organism.

    Lifecycle:
        maturing

    Examples:
        >>> human = get_experiment(census, "homo sapiens")

        >>> human = get_experiment(census, "Homo sapiens")

        >>> human = get_experiment(census, "homo_sapiens")
    """
    exp_name = _get_experiment_name(organism)

    if exp_name not in census["census_data"]:
        raise ValueError(f"Unknown organism {organism} - does not exist")
    exp = census["census_data"][exp_name]
    if exp.soma_type != "SOMAExperiment":
        raise ValueError(f"Unknown organism {organism} - not a SOMA Experiment")

    return exp



# Copyright (c) 2022, Chan Zuckerberg Initiative
#
# Licensed under the MIT License.

"""Get slice as AnnData.

Methods to retrieve slices of the census as AnnData objects.
"""

from typing import Literal, Optional, Sequence

import anndata
import pandas as pd
import tiledbsoma as soma
from somacore.options import SparseDFCoord

from ._experiment import _get_experiment, _get_experiment_name
from ._release_directory import get_census_version_directory
from ._util import _extract_census_version, _uri_join

CENSUS_EMBEDDINGS_LOCATION_BASE_URI = "s3://cellxgene-contrib-public/contrib/cell-census/soma/"


def get_anndata(
    census: soma.Collection,
    organism: str,
    measurement_name: str = "RNA",
    X_name: str = "raw",
    X_layers: Optional[Sequence[str]] = (),
    obsm_layers: Optional[Sequence[str]] = (),
    obsp_layers: Optional[Sequence[str]] = (),
    varm_layers: Optional[Sequence[str]] = (),
    varp_layers: Optional[Sequence[str]] = (),
    obs_value_filter: Optional[str] = None,
    obs_coords: Optional[SparseDFCoord] = None,
    var_value_filter: Optional[str] = None,
    var_coords: Optional[SparseDFCoord] = None,
    column_names: Optional[soma.AxisColumnNames] = None,
    obs_embeddings: Optional[Sequence[str]] = (),
    var_embeddings: Optional[Sequence[str]] = (),
) -> anndata.AnnData:
    """Convenience wrapper around :class:`tiledbsoma.Experiment` query, to build and execute a query,
    and return it as an :class:`anndata.AnnData` object.

    Args:
        census:
            The census object, usually returned by :func:`open_soma`.
        organism:
            The organism to query, usually one of ``"Homo sapiens`` or ``"Mus musculus"``.
        measurement_name:
            The measurement object to query. Defaults to ``"RNA"``.
        X_name:
            The ``X`` layer to query. Defaults to ``"raw"``.
        X_layers:
            Additional layers to add to :attr:`anndata.AnnData.layers`.
        obs_value_filter:
            Value filter for the ``obs`` metadata. Value is a filter query written in the
            SOMA ``value_filter`` syntax.
        obs_coords:
            Coordinates for the ``obs`` axis, which is indexed by the ``soma_joinid`` value.
            May be an ``int``, a list of ``int``, or a slice. The default, ``None``, selects all.
        var_value_filter:
            Value filter for the ``var`` metadata. Value is a filter query written in the
            SOMA ``value_filter`` syntax.
        var_coords:
            Coordinates for the ``var`` axis, which is indexed by the ``soma_joinid`` value.
            May be an ``int``, a list of ``int``, or a slice. The default, ``None``, selects all.
        column_names:
            Columns to fetch for ``obs`` and ``var`` dataframes.
        obsm_layers:
            Additional obsm layers to read and return in the ``obsm`` slot.
        obsp_layers:
            Additional obsp layers to read and return in the ``obsp`` slot.
        varm_layers:
            Additional varm layers to read and return in the ``varm`` slot.
        varp_layers:
            Additional varp layers to read and return in the ``varp`` slot.
        obs_embeddings:
            Additional embeddings to be returned as part of the ``obsm`` slot.
            Use :func:`get_all_available_embeddings` to retrieve available embeddings
            for this Census version and organism.
        var_embeddings:
            Additional embeddings to be returned as part of the ``varm`` slot.
            Use :func:`get_all_available_embeddings` to retrieve available embeddings
            for this Census version and organism.

    Returns:
        An :class:`anndata.AnnData` object containing the census slice.

    Lifecycle:
        experimental

    Examples:
        >>> get_anndata(census, "Mus musculus", obs_value_filter="tissue_general in ['brain', 'lung']")

        >>> get_anndata(census, "Homo sapiens", column_names={"obs": ["tissue"]})

        >>> get_anndata(census, "Homo sapiens", obs_coords=slice(0, 1000))
    """
    exp = _get_experiment(census, organism)
    obs_coords = (slice(None),) if obs_coords is None else (obs_coords,)
    var_coords = (slice(None),) if var_coords is None else (var_coords,)

    if obsm_layers and obs_embeddings and set(obsm_layers) & set(obs_embeddings):
        raise ValueError("Cannot request both `obsm_layers` and `obs_embeddings` for the same embedding name")

    if varm_layers and var_embeddings and set(varm_layers) & set(var_embeddings):
        raise ValueError("Cannot request both `varm_layers` and `var_embeddings` for the same embedding name")

    with exp.axis_query(
        measurement_name,
        obs_query=soma.AxisQuery(value_filter=obs_value_filter, coords=obs_coords),
        var_query=soma.AxisQuery(value_filter=var_value_filter, coords=var_coords),
    ) as query:
        adata = query.to_anndata(
            X_name=X_name,
            column_names=column_names,
            X_layers=X_layers,
            obsm_layers=obsm_layers,
            varm_layers=varm_layers,
            obsp_layers=obsp_layers,
            varp_layers=varp_layers,
        )

        # If obs_embeddings or var_embeddings are defined, inject them in the appropriate slot
        if obs_embeddings or var_embeddings:
            from .experimental._embedding import _get_embedding, get_embedding_metadata_by_name

            census_version = _extract_census_version(census)
            experiment_name = _get_experiment_name(organism)
            census_directory = get_census_version_directory()

            if obs_embeddings:
                obs_soma_joinids = query.obs_joinids()
                for emb in obs_embeddings:
                    emb_metadata = get_embedding_metadata_by_name(emb, experiment_name, census_version, "obs_embedding")
                    uri = _uri_join(CENSUS_EMBEDDINGS_LOCATION_BASE_URI, f"{census_version}/{emb_metadata['id']}")
                    embedding = _get_embedding(census, census_directory, census_version, uri, obs_soma_joinids)
                    adata.obsm[emb] = embedding

            if var_embeddings:
                var_soma_joinids = query.var_joinids()
                for emb in var_embeddings:
                    emb_metadata = get_embedding_metadata_by_name(emb, experiment_name, census_version, "var_embedding")
                    uri = _uri_join(CENSUS_EMBEDDINGS_LOCATION_BASE_URI, f"{census_version}/{emb_metadata['id']}")
                    embedding = _get_embedding(census, census_directory, census_version, uri, var_soma_joinids)
                    adata.varm[emb] = embedding

        return adata


def _get_axis_metadata(
    census: soma.Collection,
    axis: Literal["obs", "var"],
    organism: str,
    *,
    value_filter: Optional[str] = None,
    coords: Optional[SparseDFCoord] = slice(None),
    column_names: Optional[Sequence[str]] = None,
) -> pd.DataFrame:
    exp = _get_experiment(census, organism)
    coords = (slice(None),) if coords is None else (coords,)
    if axis == "obs":
        df = exp.obs
    elif axis == "var":
        df = exp.ms["RNA"].var
    else:
        raise ValueError(f"axis should be either 'obs' or 'var', but '{axis}' was passed")
    result: pd.DataFrame = (
        df.read(coords=coords, column_names=column_names, value_filter=value_filter).concat().to_pandas()
    )
    return result


def get_obs(
    census: soma.Collection,
    organism: str,
    *,
    value_filter: Optional[str] = None,
    coords: Optional[SparseDFCoord] = slice(None),
    column_names: Optional[Sequence[str]] = None,
) -> pd.DataFrame:
    """Get the observation metadata for a query on the census.

    Args:
        census:
            The census object, usually returned by :func:`open_soma`.
        organism:
            The organism to query, usually one of ``"Homo sapiens`` or ``"Mus musculus"``
        value_filter:
            Value filter for the ``obs`` metadata. Value is a filter query written in the
            SOMA ``value_filter`` syntax.
        coords:
            Coordinates for the ``obs`` axis, which is indexed by the ``soma_joinid`` value.
            May be an ``int``, a list of ``int``, or a slice. The default, ``None``, selects all.
        column_names:
            Columns to fetch.

    Returns:
        A :class:`pandas.DataFrame` object containing metadata for the queried slice.
    """
    return _get_axis_metadata(
        census, "obs", organism, value_filter=value_filter, coords=coords, column_names=column_names
    )


def get_var(
    census: soma.Collection,
    organism: str,
    *,
    value_filter: Optional[str] = None,
    coords: Optional[SparseDFCoord] = slice(None),
    column_names: Optional[Sequence[str]] = None,
) -> pd.DataFrame:
    """Get the variable metadata for a query on the census.

    Args:
        census:
            The census object, usually returned by :func:`open_soma`.
        organism:
            The organism to query, usually one of ``"Homo sapiens`` or ``"Mus musculus"``
        value_filter:
            Value filter for the ``var`` metadata. Value is a filter query written in the
            SOMA ``value_filter`` syntax.
        coords:
            Coordinates for the ``var`` axis, which is indexed by the ``soma_joinid`` value.
            May be an ``int``, a list of ``int``, or a slice. The default, ``None``, selects all.
        column_names:
            Columns to fetch.

    Returns:
        A :class:`pandas.DataFrame` object containing metadata for the queried slice.
    """
    return _get_axis_metadata(
        census, "var", organism, value_filter=value_filter, coords=coords, column_names=column_names
    )
